@@ -12,7 +12,7 @@ use AnyEvent::WebSocket::Client;
 use AnyEvent::Future qw(as_future_cb);
 
 use vars qw<$VERSION $magic @CARP_NOT>;
-$VERSION = '0.01';
+$VERSION = '0.02';
 
 =head1 SYNOPSIS
 
@@ -84,4 +84,19 @@ sub future {
     AnyEvent::Future->new
 }
 
+=head2 C<< $transport->sleep( $seconds ) >>
+
+    $transport->sleep( 10 )->get; # wait for 10 seconds
+
+Returns a Future that will be resolved in the number of seconds given.
+
+=cut
+
+sub sleep( $self, $seconds ) {
+
+    my $res = as_future_cb( sub( $done_cb, $fail_cb ) {
+        AnyEvent->timer( after => $seconds, cb => $done_cb )
+    });
+}
+
 1;
\ No newline at end of file
